from mage_ai.io.config import ConfigFileLoader, ConfigKey, EnvironmentVariableLoader
from mage_ai.tests.base_test import DBTestCase
from pathlib import Path
from unittest import mock


class ConfigLoaderTests(DBTestCase):
    def setUp(self):
        super().setUp()
        self.test_path = Path('./test')
        self.test_config_path = self.test_path / 'io_config.yaml'
        self.test_config_path_verbose = self.test_path / 'old_io_config.yaml'
        sample_yaml = """default:
  AWS_REGION: test_region
  GOOGLE_SERVICE_ACC_KEY_FILEPATH: "path/to/test/key.json"
  POSTGRES_DBNAME: my_psql_db
  REDSHIFT_TEMP_CRED_PASSWORD: a_strong_password
  REDSHIFT_PORT: 5439
  SNOWFLAKE_DEFAULT_SCHEMA: sample_schema
contains_check:
  AWS_REGION: region_test
  GOOGLE_SERVICE_ACC_KEY_FILEPATH: "a/path/to/nowhere"
  REDSHIFT_TEMP_CRED_PASSWORD: a_strong_password
  REDSHIFT_PORT: 5439
  SNOWFLAKE_DEFAULT_SCHEMA: sample_schema
a_diff_profile:
  AWS_REGION: region_test
  GOOGLE_SERVICE_ACC_KEY_FILEPATH: "a/path/to/nowhere"
  POSTGRES_DBNAME: another_psql_db
  REDSHIFT_TEMP_CRED_PASSWORD: another_strong_password
  REDSHIFT_PORT: 9453
  SNOWFLAKE_DEFAULT_SCHEMA: schema_two
template:
  REDSHIFT_CLUSTER_ID: "{{ env_var('REDSHIFT_CLUSTER') }}"
"""
        sample_yaml_verbose_format = """default:
  #  Default profile created for data IO access. Add credentials for the sources you use and
  #   remove the rest.
  BigQuery:
    credentials_mapping:
        test: 4
  AWS:
    Redshift:
      database: your_redshift_database_name
      port: your_redshift_cluster_port
    region: your_aws_region
  PostgreSQL:
    database: your_postgres_database_name
"""
        with self.test_config_path.open('w') as fout:
            fout.write(sample_yaml)
        with self.test_config_path_verbose.open('w') as fout:
            fout.write(sample_yaml_verbose_format)

    def tearDown(self):
        self.test_config_path.unlink()
        self.test_config_path_verbose.unlink()
        super().tearDown()

    def test_config_map_contains(self):
        expected_keys = [
            ConfigKey.AWS_REGION,
            ConfigKey.GOOGLE_SERVICE_ACC_KEY,
            ConfigKey.REDSHIFT_TEMP_CRED_PASSWORD,
            ConfigKey.REDSHIFT_DBUSER,
            ConfigKey.POSTGRES_DBNAME,
            ConfigKey.SNOWFLAKE_DEFAULT_SCHEMA,
            ConfigKey.REDSHIFT_CLUSTER_ID,
        ]
        default_expected_values = [True, False, True, False, False, True, False]

        config = ConfigFileLoader(self.test_config_path, profile='contains_check')
        for expected_key, expected_value in zip(expected_keys, default_expected_values):
            self.assertTrue((expected_key in config) == expected_value)

    def test_config_map_get(self):
        expected_keys = [
            ConfigKey.AWS_REGION.value,
            ConfigKey.GOOGLE_SERVICE_ACC_KEY_FILEPATH.value,
            ConfigKey.REDSHIFT_TEMP_CRED_PASSWORD.value,
            ConfigKey.REDSHIFT_PORT.value,
            ConfigKey.POSTGRES_DBNAME.value,
            ConfigKey.SNOWFLAKE_DEFAULT_SCHEMA.value,
        ]
        default_expected_values = [
            'test_region',
            'path/to/test/key.json',
            'a_strong_password',
            5439,
            'my_psql_db',
            'sample_schema',
        ]
        diff_expected_values = [
            'region_test',
            'a/path/to/nowhere',
            'another_strong_password',
            9453,
            'another_psql_db',
            'schema_two',
        ]

        expected_keys.append(ConfigKey.REDSHIFT_CLUSTER_ID)
        default_expected_values.append(None)
        diff_expected_values.append(None)
        config = ConfigFileLoader(self.test_config_path, profile='default')
        for expected_key, expected_value in zip(expected_keys, default_expected_values):
            self.assertEqual(config[expected_key], expected_value)
        config = ConfigFileLoader(self.test_config_path, profile='a_diff_profile')
        for expected_key, expected_value in zip(expected_keys, diff_expected_values):
            self.assertEqual(config[expected_key], expected_value)

    def test_config_map_get_old(self):
        expected_keys = [
            ConfigKey.AWS_REGION,
            ConfigKey.GOOGLE_SERVICE_ACC_KEY_FILEPATH,
            ConfigKey.REDSHIFT_CLUSTER_ID,
            ConfigKey.REDSHIFT_PORT,
            ConfigKey.POSTGRES_DBNAME,
            ConfigKey.SNOWFLAKE_DEFAULT_SCHEMA,
            'a bad key',
        ]
        default_expected_values = [
            'your_aws_region',
            None,
            None,
            'your_redshift_cluster_port',
            'your_postgres_database_name',
            None,
            None,
        ]

        config = ConfigFileLoader(self.test_config_path_verbose, profile='default')
        for expected_key, expected_value in zip(expected_keys, default_expected_values):
            self.assertEqual(config[expected_key], expected_value)

    @mock.patch('mage_ai.io.config.os')
    def test_env_map_contains(self, mock_os):
        expected_keys = [
            ConfigKey.AWS_SECRET_ACCESS_KEY,
            ConfigKey.GOOGLE_SERVICE_ACC_KEY_FILEPATH,
            ConfigKey.REDSHIFT_DBNAME,
            ConfigKey.POSTGRES_HOST,
            ConfigKey.SNOWFLAKE_PASSWORD,
        ]
        values = [
            'aws_secret_access_key',
            'filepath',
            'test_db',
            'url_to_db',
            'a_snowflake_password',
        ]

        test_env_vars = dict(zip(expected_keys, values))
        mock_os.environ = test_env_vars

        expected_keys.append(ConfigKey.REDSHIFT_CLUSTER_ID)
        expected_keys.append(ConfigKey.SNOWFLAKE_ACCOUNT)
        expected_keys.append(ConfigKey.POSTGRES_USER)
        expected_keys.append(ConfigKey.POSTGRES_DBNAME)

        expected_values = [True] * 5 + [False] * 4

        env_loader = EnvironmentVariableLoader()
        for expected_key, expected_value in zip(expected_keys, expected_values):
            self.assertTrue((expected_key in env_loader) == expected_value)

    @mock.patch('mage_ai.io.config.os')
    def test_env_map_get(self, mock_os):
        expected_keys = [
            ConfigKey.AWS_SECRET_ACCESS_KEY,
            ConfigKey.GOOGLE_SERVICE_ACC_KEY_FILEPATH,
            ConfigKey.REDSHIFT_DBNAME,
            ConfigKey.POSTGRES_HOST,
            ConfigKey.SNOWFLAKE_PASSWORD,
        ]
        expected_values = [
            'aws_secret_access_key',
            'filepath',
            'test_db',
            'url_to_db',
            'a_snowflake_password',
        ]

        test_env_vars = dict(zip(expected_keys, expected_values))
        mock_os.getenv = test_env_vars.get

        expected_keys.append(ConfigKey.REDSHIFT_CLUSTER_ID)
        expected_values.append(None)

        env_loader = EnvironmentVariableLoader()
        for expected_key, expected_value in zip(expected_keys, expected_values):
            self.assertEqual(env_loader[expected_key], expected_value)

    @mock.patch('mage_ai.data_preparation.shared.utils.os')
    def test_env_map_get_redshift(self, mock_os):
        test_env_vars = {'REDSHIFT_CLUSTER': 'env_var_cluster'}
        mock_os.getenv = test_env_vars.get
        config = ConfigFileLoader(self.test_config_path, profile='template')
        self.assertEqual(config[ConfigKey.REDSHIFT_CLUSTER_ID], 'env_var_cluster')
        self.assertEqual(config[ConfigKey.REDSHIFT_DBUSER], None)



from keyword import iskeyword
from mage_ai.data_cleaner.column_types.column_type_detector import REGEX_NUMBER, infer_column_types
from mage_ai.data_cleaner.column_types.constants import ColumnType
from mage_ai.data_cleaner.transformer_actions.constants import (
    ActionType,
    Axis,
    NameConventionPatterns,
)
from pandas import DataFrame
from typing import Dict, List, Union
import logging

logger = logging.getLogger(__name__)


def build_action_variables(
    df: DataFrame, ctypes: Dict[str, ColumnType] = None, columns: List[str] = None
) -> Dict:
    """
    Builds action variable set from data frame for use with transformer actions library.

    Args:
        df (DataFrame): Data frame to build transformer actions for.
        ctypes (Dict[str, ColumnType], optional): Column type for each column in data frame.
        Defaults to None, in which case the column types will be automatically inferred.
        columns (List[str], options): Columns to generate action variables for. Defaults to None,
        in which case action variables are generated for all columns of the data frame.
    Returns:
        Dict: Set of action variables.
    """
    if ctypes is None:
        ctypes = {}
    if columns is None:
        columns = df.columns
    ctypes = infer_column_types(df, column_types=ctypes)
    variable_set = {}
    for column_name in columns:
        variable_set[column_name] = {
            'feature': {
                'column_type': ctypes[column_name],
                'uuid': column_name,
            },
            'type': 'feature',
        }
    return variable_set


def build_transformer_action(
    df: DataFrame,
    action_type: Union[ActionType, str],
    arguments: List[str] = [],
    action_code: str = '',
    options: Dict = {},
    axis: Union[Axis, str] = Axis.COLUMN,
    outputs: List[Dict] = [],
) -> Dict:
    """
    Builds transformer action payload from arguments. The output of this function can be passed
    as input to the `transformer_actions.base.BaseAction` to perform the requested transformation.

    Designed as a helper method to simplify generating transformer action payloads

    Args:
        df (DataFrame): The data frame to build a transformer action payload for.
        action_type (Union[ActionType, str]): Transformer action to perform.
        arguments (List[str], optional): Columns/Rows to perform this action on.
        Defaults to [].
        action_code (str, optional): Special code or query to execute with action. Defaults to ''.
        options (Dict, optional): Options specifying behavior of action. Defaults to {}.
        axis (Union[Axis, str], optional): Axis of the data frame to apply the action to.
        Defaults to Axis.COLUMN.
        outputs (List[Dict], optional): Specifies metadata of newly created columns.
        Defaults to [].

    Returns:
        Dict: Transformer action payload
    """
    action_variables = build_action_variables(df)
    return dict(
        action_type=action_type,
        action_arguments=list(arguments),
        action_code=action_code,
        action_options=options,
        action_variables=action_variables,
        axis=axis,
        outputs=outputs,
    )


def clean_column_name(name):
    if iskeyword(name):
        name = f'{name}_'
    name = name.strip(' \'\"_-')
    name = NameConventionPatterns.CONNECTORS.sub('_', name)
    name = NameConventionPatterns.NON_ALNUM.sub('', name)
    name = REGEX_NUMBER.sub(lambda number: f'number_{number.group(0)}', name)
    if iskeyword(name):
        name = f'{name}_'
    uppercase_group = NameConventionPatterns.UPPERCASE.match(name)
    pascal_group = NameConventionPatterns.PASCAL.match(name)
    camel_group = NameConventionPatterns.CAMEL.match(name)
    if uppercase_group:
        name = name.lower()
    elif pascal_group:
        components = NameConventionPatterns.PASCAL_COMPONENT.findall(name)
        name = '_'.join(components)
    elif camel_group:
        components = NameConventionPatterns.CAMEL_COMPONENT.findall(name)
        components += NameConventionPatterns.PASCAL_COMPONENT.findall(name)
        name = '_'.join(components)
    return name.lower()


def columns_to_remove(transformer_actions):
    arr = filter(
        lambda x: x['action_type'] == ActionType.REMOVE and x['axis'] == Axis.COLUMN,
        transformer_actions,
    )

    columns = []
    for transformer_action in arr:
        columns += transformer_action['action_arguments']

    return columns


def generate_action_titles(transformer_actions):
    for a in transformer_actions:
        if not a.get('title'):
            a['title'] = generate_action_title(a)
    return transformer_actions


def generate_action_title(action):
    payload = action['action_payload']
    action_type = payload['action_type']
    title = ''
    if action_type == ActionType.REMOVE:
        title = 'Remove columns'
    elif action_type == ActionType.FILTER:
        title = 'Filter rows'
    elif action_type == ActionType.DROP_DUPLICATE:
        title = 'Drop duplicate rows'
    elif action_type == ActionType.REFORMAT:
        title = 'Reformat values'
    elif action_type == ActionType.IMPUTE:
        title = 'Fill in missing values'
    elif action_type == ActionType.CLEAN_COLUMN_NAME:
        title = 'Clean dirty column names'
    return title


def generate_string_cols(df, columns):
    for column in columns:
        clean_col = df[column]
        dropped = clean_col.dropna(axis=0)
        exact_dtype = type(dropped.iloc[0]) if len(dropped) > 0 else None
        if exact_dtype is str:
            yield column
        else:
            logger.warn(
                f'Attempted to perform string-only action on non-string column \'{column}\''
            )



from mage_ai.data_cleaner.transformer_actions.udf.base import BaseUDF


class Substring(BaseUDF):
    def execute(self):
        start = self.options.get('start')
        stop = self.options.get('stop')
        if start is None and stop is None:
            raise Exception('Require at least one of `start` and `stop` parameters.')
        return self.df[self.arguments[0]].str.slice(start=start, stop=stop)



import json
import os
import traceback
from enum import Enum
from typing import Any, Dict, List

import numpy as np
import pandas as pd
import polars as pl
from pandas.api.types import is_object_dtype
from pandas.core.indexes.range import RangeIndex

from mage_ai.data_cleaner.shared.utils import is_geo_dataframe, is_spark_dataframe
from mage_ai.data_preparation.models.constants import (
    DATAFRAME_ANALYSIS_KEYS,
    DATAFRAME_SAMPLE_COUNT,
    DATAFRAME_SAMPLE_MAX_COLUMNS,
    VARIABLE_DIR,
)
from mage_ai.data_preparation.models.utils import (  # dask_from_pandas,
    STRING_SERIALIZABLE_COLUMN_TYPES,
    apply_transform_pandas,
    cast_column_types,
    deserialize_columns,
    serialize_columns,
)
from mage_ai.data_preparation.storage.base_storage import BaseStorage
from mage_ai.data_preparation.storage.local_storage import LocalStorage
from mage_ai.shared.parsers import sample_output
from mage_ai.shared.utils import clean_name

DATAFRAME_COLUMN_TYPES_FILE = 'data_column_types.json'
DATAFRAME_PARQUET_FILE = 'data.parquet'
DATAFRAME_PARQUET_SAMPLE_FILE = 'sample_data.parquet'
DATAFRAME_CSV_FILE = 'data.csv'

JSON_FILE = 'data.json'
JSON_SAMPLE_FILE = 'sample_data.json'


class VariableType(str, Enum):
    DATAFRAME = 'dataframe'
    DATAFRAME_ANALYSIS = 'dataframe_analysis'
    GEO_DATAFRAME = 'geo_dataframe'
    POLARS_DATAFRAME = 'polars_dataframe'
    SPARK_DATAFRAME = 'spark_dataframe'


class Variable:
    def __init__(
        self,
        uuid: str,
        pipeline_path: str,
        block_uuid: str,
        partition: str = None,
        spark=None,
        storage: BaseStorage = None,
        variable_type: VariableType = None
    ) -> None:
        self.uuid = uuid
        if storage is None:
            self.storage = LocalStorage()
        else:
            self.storage = storage
        # if not self.storage.path_exists(pipeline_path):
        #     raise Exception(f'Pipeline path {pipeline_path} does not exist.')
        self.pipeline_path = pipeline_path
        self.block_uuid = block_uuid
        self.block_dir_name = clean_name(self.block_uuid)
        self.partition = partition
        self.variable_dir_path = os.path.join(
            pipeline_path,
            VARIABLE_DIR,
            partition or '',
            self.block_dir_name,
        )
        if not self.storage.path_exists(self.variable_dir_path):
            self.storage.makedirs(self.variable_dir_path)

        self.variable_type = variable_type
        self.check_variable_type(spark=spark)

    @property
    def variable_path(self):
        return os.path.join(self.variable_dir_path, f'{self.uuid}')

    @classmethod
    def dir_path(self, pipeline_path, block_uuid):
        return os.path.join(pipeline_path, VARIABLE_DIR, clean_name(block_uuid))

    def check_variable_type(self, spark=None):
        """
        Infer variable type based on data in the storage.
        """
        if self.variable_type is None and self.storage.path_exists(
            os.path.join(self.variable_path, DATAFRAME_PARQUET_FILE)
        ):
            # If parquet file exists for given variable, set the variable type to DATAFRAME
            self.variable_type = VariableType.DATAFRAME
        elif ((self.variable_type == VariableType.DATAFRAME or self.variable_type is None)
                and os.path.exists(
                os.path.join(self.variable_dir_path, f'{self.uuid}', 'data.sh'))):
            self.variable_type = VariableType.GEO_DATAFRAME
        elif self.variable_type is None and \
                len(self.storage.listdir(self.variable_path, suffix='.parquet')) > 0 and \
                spark is not None:
            self.variable_type = VariableType.SPARK_DATAFRAME

    def convert_parquet_to_csv(self):
        """
        For DATAFRAME variable, convert parquet files to csv files.
        """
        if self.variable_type != VariableType.DATAFRAME:
            return
        csv_file_path = os.path.join(self.variable_path, DATAFRAME_CSV_FILE)
        if self.storage.path_exists(csv_file_path):
            return
        df = self.__read_parquet()
        self.storage.write_csv(df, csv_file_path)

    def delete(self):
        """
        Delete the variable data.
        """
        if self.variable_type is None and self.storage.path_exists(
            os.path.join(self.variable_dir_path, f'{self.uuid}', DATAFRAME_PARQUET_FILE)
        ):
            # If parquet file exists for given variable, set the variable type to DATAFRAME
            self.variable_type = VariableType.DATAFRAME
        if self.variable_type == VariableType.DATAFRAME:
            self.__delete_parquet()
        elif self.variable_type == VariableType.DATAFRAME_ANALYSIS:
            return self.__delete_dataframe_analysis()
        return self.__delete_json()

    def read_data(
        self,
        dataframe_analysis_keys: List[str] = None,
        sample: bool = False,
        sample_count: int = None,
        spark=None,
    ) -> Any:
        """
        Read variable data.

        Args:
            dataframe_analysis_keys (List[str], optional): For DATAFRAME_ANALYSIS variable,
                only read the selected keys.
            sample (bool, optional): Whether to sample the rows of a dataframe, used for
                DATAFRAME variable.
            sample_count (int, optional): The number of rows to sample, used for
                DATAFRAME variable.
            spark (None, optional): Spark context, used to read SPARK_DATAFRAME variable.
        """
        if self.variable_type == VariableType.DATAFRAME:
            return self.__read_parquet(sample=sample, sample_count=sample_count)
        elif self.variable_type == VariableType.SPARK_DATAFRAME:
            return self.__read_spark_parquet(sample=sample, sample_count=sample_count, spark=spark)
        elif self.variable_type == VariableType.GEO_DATAFRAME:
            return self.__read_geo_dataframe(sample=sample, sample_count=sample_count)
        elif self.variable_type == VariableType.DATAFRAME_ANALYSIS:
            return self.__read_dataframe_analysis(dataframe_analysis_keys=dataframe_analysis_keys)
        return self.__read_json(sample=sample)

    async def read_data_async(
        self,
        dataframe_analysis_keys: List[str] = None,
        sample: bool = False,
        sample_count: int = None,
        spark=None,
    ):
        """
        Read variable data asynchronously.

        Args:
            dataframe_analysis_keys (List[str], optional): For DATAFRAME_ANALYSIS variable,
                only read the selected keys.
            sample (bool, optional): Whether to sample the rows of a dataframe, used for
                DATAFRAME variable.
            sample_count (int, optional): The number of rows to sample, used for
                DATAFRAME variable.
            spark (None, optional): Spark context, used to read SPARK_DATAFRAME variable.
        """
        if self.variable_type == VariableType.DATAFRAME:
            return self.__read_parquet(sample=sample, sample_count=sample_count)
        elif self.variable_type == VariableType.SPARK_DATAFRAME:
            return self.__read_spark_parquet(sample=sample, sample_count=sample_count, spark=spark)
        elif self.variable_type == VariableType.DATAFRAME_ANALYSIS:
            return await self.__read_dataframe_analysis_async(
                dataframe_analysis_keys=dataframe_analysis_keys,
            )
        return await self.__read_json_async(sample=sample)

    def write_data(self, data: Any) -> None:
        """
        Write variable data to the persistent storage.

        Args:
            data (Any): Variable data to be written to storage.
        """
        if self.variable_type is None and type(data) is pd.DataFrame:
            self.variable_type = VariableType.DATAFRAME
        elif self.variable_type is None and type(data) is pl.DataFrame:
            self.variable_type = VariableType.POLARS_DATAFRAME
        elif is_spark_dataframe(data):
            self.variable_type = VariableType.SPARK_DATAFRAME
        elif is_geo_dataframe(data):
            self.variable_type = VariableType.GEO_DATAFRAME

        if self.variable_type == VariableType.DATAFRAME:
            self.__write_parquet(data)
        elif self.variable_type == VariableType.POLARS_DATAFRAME:
            self.__write_polars_dataframe(data)
        elif self.variable_type == VariableType.SPARK_DATAFRAME:
            self.__write_spark_parquet(data)
        elif self.variable_type == VariableType.GEO_DATAFRAME:
            self.__write_geo_dataframe(data)
        elif self.variable_type == VariableType.DATAFRAME_ANALYSIS:
            self.__write_dataframe_analysis(data)
        else:
            self.__write_json(data)

    async def write_data_async(self, data: Any) -> None:
        """
        Write variable data to the persistent storage.

        Args:
            data (Any): Variable data to be written to storage.
        """
        if self.variable_type is None and type(data) is pd.DataFrame:
            self.variable_type = VariableType.DATAFRAME
        elif self.variable_type is None and type(data) is pl.DataFrame:
            self.variable_type = VariableType.POLARS_DATAFRAME
        elif is_spark_dataframe(data):
            self.variable_type = VariableType.SPARK_DATAFRAME
        elif is_geo_dataframe(data):
            self.variable_type = VariableType.GEO_DATAFRAME

        if self.variable_type == VariableType.DATAFRAME:
            self.__write_parquet(data)
        elif self.variable_type == VariableType.POLARS_DATAFRAME:
            self.__write_polars_dataframe(data)
        elif self.variable_type == VariableType.SPARK_DATAFRAME:
            self.__write_spark_parquet(data)
        elif self.variable_type == VariableType.GEO_DATAFRAME:
            self.__write_geo_dataframe(data)
        elif self.variable_type == VariableType.DATAFRAME_ANALYSIS:
            self.__write_dataframe_analysis(data)
        else:
            await self.__write_json_async(data)

    def __delete_dataframe_analysis(self) -> None:
        for k in DATAFRAME_ANALYSIS_KEYS:
            file_path = os.path.join(self.variable_path, f'{k}.json')
            if self.storage.path_exists(file_path):
                try:
                    self.storage.remove(file_path)
                except FileNotFoundError as err:
                    print(f'Error deleting file {file_path}: {err}')

    def __delete_json(self) -> None:
        old_file_path = os.path.join(self.variable_dir_path, f'{self.uuid}.json')
        if self.storage.path_exists(old_file_path):
            self.storage.remove(old_file_path)
        if self.storage.isdir(self.variable_path):
            self.storage.remove_dir(self.variable_path)

    def __delete_parquet(self) -> None:
        file_path = os.path.join(self.variable_path, DATAFRAME_PARQUET_FILE)

        if self.storage.path_exists(file_path):
            self.storage.remove(file_path)
            self.storage.remove_dir(self.variable_path)

    def __read_json(self, default_value: Dict = None, sample: bool = False) -> Dict:
        if default_value is None:
            default_value = {}
        # For backward compatibility
        old_file_path = os.path.join(self.variable_dir_path, f'{self.uuid}.json')
        file_path = os.path.join(self.variable_path, JSON_FILE)
        sample_file_path = os.path.join(self.variable_path, JSON_SAMPLE_FILE)

        read_sample_success = False
        if sample and self.storage.path_exists(sample_file_path):
            try:
                data = self.storage.read_json_file(sample_file_path, default_value)
                read_sample_success = True
            except Exception:
                pass
        if not read_sample_success:
            if self.storage.path_exists(file_path):
                data = self.storage.read_json_file(file_path, default_value)
            else:
                data = self.storage.read_json_file(old_file_path, default_value)
        if sample:
            data = sample_output(data)[0]
        return data

    async def __read_json_async(self, default_value: Dict = None, sample: bool = False) -> Dict:
        if default_value is None:
            default_value = {}
        # For backward compatibility
        old_file_path = os.path.join(self.variable_dir_path, f'{self.uuid}.json')
        file_path = os.path.join(self.variable_path, JSON_FILE)
        sample_file_path = os.path.join(self.variable_path, JSON_SAMPLE_FILE)

        read_sample_success = False
        if sample and self.storage.path_exists(sample_file_path):
            try:
                data = await self.storage.read_json_file_async(sample_file_path, default_value)
                read_sample_success = True
            except Exception:
                pass
        if not read_sample_success:
            if self.storage.path_exists(file_path):
                data = await self.storage.read_json_file_async(file_path, default_value)
            else:
                data = await self.storage.read_json_file_async(old_file_path, default_value)
        if sample:
            data = sample_output(data)[0]
        return data

    def __write_json(self, data) -> None:
        if not self.storage.isdir(self.variable_path):
            self.storage.makedirs(self.variable_path, exist_ok=True)
        file_path = os.path.join(self.variable_path, JSON_FILE)
        sample_file_path = os.path.join(self.variable_path, JSON_SAMPLE_FILE)
        self.storage.write_json_file(file_path, data)
        self.storage.write_json_file(sample_file_path, sample_output(data)[0])

    async def __write_json_async(self, data) -> None:
        if not self.storage.isdir(self.variable_path):
            self.storage.makedirs(self.variable_path, exist_ok=True)
        file_path = os.path.join(self.variable_path, JSON_FILE)
        sample_file_path = os.path.join(self.variable_path, JSON_SAMPLE_FILE)
        try:
            await self.storage.write_json_file_async(file_path, data)
            await self.storage.write_json_file_async(sample_file_path, sample_output(data)[0])
        except Exception:
            traceback.print_exc()

    def __read_geo_dataframe(self, sample: bool = False, sample_count: int = None):
        import geopandas as gpd

        file_path = os.path.join(self.variable_path, 'data.sh')
        sample_file_path = os.path.join(self.variable_path, 'sample_data.sh')
        if not os.path.exists(file_path):
            return gpd.GeoDataFrame()
        if sample and os.path.exists(sample_file_path):
            try:
                df = gpd.read_file(sample_file_path)
            except Exception:
                df = gpd.read_file(file_path)
        else:
            df = gpd.read_file(file_path)
        if sample:
            sample_count = sample_count or DATAFRAME_SAMPLE_COUNT
            if df.shape[0] > sample_count:
                df = df.iloc[:sample_count]
        return df

    def __read_parquet(
        self,
        sample: bool = False,
        sample_count: int = None,
        raise_exception: bool = False,
    ) -> pd.DataFrame:
        file_path = os.path.join(self.variable_path, DATAFRAME_PARQUET_FILE)
        sample_file_path = os.path.join(self.variable_path, DATAFRAME_PARQUET_SAMPLE_FILE)

        read_sample_success = False
        if sample:
            try:
                df = self.storage.read_parquet(sample_file_path, engine='pyarrow')
                read_sample_success = True
            except Exception as e:
                if raise_exception:
                    raise e
                pass
        if not read_sample_success:
            try:
                df = self.storage.read_parquet(file_path, engine='pyarrow')
            except Exception as e:
                if raise_exception:
                    raise e
                df = pd.DataFrame()
        if sample:
            sample_count = sample_count or DATAFRAME_SAMPLE_COUNT
            if df.shape[0] > sample_count:
                df = df.iloc[:sample_count]

        column_types_filename = os.path.join(self.variable_path, DATAFRAME_COLUMN_TYPES_FILE)
        if os.path.exists(column_types_filename):
            with open(column_types_filename, 'r') as f:
                column_types = json.load(f)
                # ddf = dask_from_pandas(df)
                df = apply_transform_pandas(df, lambda row: deserialize_columns(row, column_types))
                df = cast_column_types(df, column_types)
        return df

    def __read_spark_parquet(self, sample: bool = False, sample_count: int = None, spark=None):
        if spark is None:
            return None
        df = (
            spark.read
            .format('parquet')
            .option('header', 'true')
            .option('inferSchema', 'true')
            .option('delimiter', ',')
            .load(self.variable_path)
        )
        if sample and sample_count:
            df = df.limit(sample_count)
        return df

    def __write_geo_dataframe(self, data) -> None:
        os.makedirs(self.variable_path, exist_ok=True)
        data.to_file(os.path.join(self.variable_path, 'data.sh'))
        df_sample_output = data.iloc[:DATAFRAME_SAMPLE_COUNT]
        df_sample_output.to_file(os.path.join(self.variable_path, 'sample_data.sh'))

    def __write_parquet(self, data: pd.DataFrame) -> None:
        column_types = {}
        df_output = data.copy()
        # Clean up data types since parquet doesn't support mixed data types
        for c in df_output.columns:
            c_dtype = df_output[c].dtype
            if not is_object_dtype(c_dtype):
                column_types[c] = str(c_dtype)
            else:
                series_non_null = df_output[c].dropna()
                if len(series_non_null) > 0:
                    coltype = type(series_non_null.iloc[0])
                    if is_object_dtype(series_non_null.dtype):
                        if coltype.__name__ in STRING_SERIALIZABLE_COLUMN_TYPES:
                            cast_coltype = str
                        else:
                            cast_coltype = coltype
                        try:
                            df_output[c] = series_non_null.astype(cast_coltype)
                            coltype = str
                        except Exception:
                            # Fall back to convert to string
                            # df_output[c] = series_non_null.astype(str)
                            pass

                    col_not_numpy_type = coltype.__module__ != np.__name__
                    col_is_numpy_array_type = coltype.__name__ == np.ndarray.__name__
                    if col_not_numpy_type or col_is_numpy_array_type:
                        column_types[c] = coltype.__name__
                    else:
                        column_types[c] = type(series_non_null.iloc[0].item()).__name__

        self.storage.makedirs(self.variable_path, exist_ok=True)
        with open(os.path.join(self.variable_path, DATAFRAME_COLUMN_TYPES_FILE), 'w') as f:
            f.write(json.dumps(column_types))

        # Try using Polars to write the dataframe to improve performance
        if type(df_output.index) is RangeIndex and df_output.index.start == 0 \
                and df_output.index.stop == df_output.shape[0] and df_output.index.step == 1:
            # Polars ignores any index
            try:
                pl_df = pl.from_pandas(df_output)
                self.__write_polars_dataframe(pl_df)
                # Test read dataframe from parquet
                self. __read_parquet(sample=True, raise_exception=True)

                return
            except Exception:
                pass

        # ddf = dask_from_pandas(df_output)
        df_output_serialized = apply_transform_pandas(
            df_output,
            lambda row: serialize_columns(row, column_types),
        )

        self.storage.write_parquet(
            df_output_serialized,
            os.path.join(self.variable_path, DATAFRAME_PARQUET_FILE),
        )

        try:
            df_sample_output = df_output_serialized.iloc[
                :DATAFRAME_SAMPLE_COUNT,
                :DATAFRAME_SAMPLE_MAX_COLUMNS
            ]

            self.storage.write_parquet(
                df_sample_output,
                os.path.join(self.variable_path, DATAFRAME_PARQUET_SAMPLE_FILE),
            )
        except Exception as err:
            print(f'Sample output error: {err}.')
            traceback.print_exc()

    def __write_polars_dataframe(self, data: pl.DataFrame) -> None:
        self.storage.makedirs(self.variable_path, exist_ok=True)

        self.storage.write_polars_dataframe(
            data,
            os.path.join(self.variable_path, DATAFRAME_PARQUET_FILE),
        )

        try:
            sample_columns = data.columns[:DATAFRAME_SAMPLE_MAX_COLUMNS]
            df_sample_output = data[
                :DATAFRAME_SAMPLE_COUNT,
                sample_columns,
            ]

            self.storage.write_polars_dataframe(
                df_sample_output,
                os.path.join(self.variable_path, DATAFRAME_PARQUET_SAMPLE_FILE),
            )
        except Exception as err:
            print(f'Sample output error: {err}.')
            traceback.print_exc()

    def __write_spark_parquet(self, data) -> None:
        (
            data.write
            .option('header', 'True')
            .mode('overwrite')
            .parquet(self.variable_path)
        )

    def __read_dataframe_analysis(
        self,
        dataframe_analysis_keys: List[str] = None,
    ) -> Dict[str, Dict]:
        """
        Read the following files
        1. metadata.json
        2. statistics.json
        3. insights.json
        4. suggestions.json
        """
        if not self.storage.path_exists(self.variable_path):
            return dict()
        result = dict()
        for k in DATAFRAME_ANALYSIS_KEYS:
            if dataframe_analysis_keys is not None and k not in dataframe_analysis_keys:
                continue
            result[k] = self.storage.read_json_file(os.path.join(self.variable_path, f'{k}.json'))
        return result

    async def __read_dataframe_analysis_async(
        self,
        dataframe_analysis_keys: List[str] = None,
    ) -> Dict[str, Dict]:
        """
        Read the following files
        1. metadata.json
        2. statistics.json
        3. insights.json
        4. suggestions.json
        """
        if not self.storage.path_exists(self.variable_path):
            return dict()
        result = dict()
        for k in DATAFRAME_ANALYSIS_KEYS:
            if dataframe_analysis_keys is not None and k not in dataframe_analysis_keys:
                continue
            result[k] = await self.storage.read_json_file_async(
                os.path.join(self.variable_path, f'{k}.json'))
        return result

    def __write_dataframe_analysis(self, data: Dict[str, Dict]) -> None:
        """
        Write the following files
        1. metadata.json
        2. statistics.json
        3. insights.json
        4. suggestions.json
        """
        self.storage.makedirs(self.variable_path, exist_ok=True)
        for k in DATAFRAME_ANALYSIS_KEYS:
            self.storage.write_json_file(os.path.join(self.variable_path, f'{k}.json'), data.get(k))



from datetime import datetime
from mage_ai.services.metaplane.config import Config
from mage_ai.shared.http_client import HttpClient
from typing import Dict, List, Literal, TypedDict, Union
import dateutil.parser
import time


class Connection(TypedDict):
    createdAt: datetime
    id: str
    isEnabled: bool
    name: str
    status: Literal['active', 'deleted']
    updatedAt: datetime


class Monitor(TypedDict):
    createdAt: datetime
    cronTab: str
    id: str
    isEnabled: bool
    type: Literal[
        'cardinality',
        'column_count',
        'custom',
        'duration',
        'freshness',
        'iqr',
        'kurtosis',
        'lower_percentile',
        'max',
        'mean',
        'median',
        'min',
        'monotonicity',
        'nullness',
        'range',
        'row_count',
        'skew',
        'stddev',
        'uniqueness',
        'upper_percentile',
    ]
    updatedAt: datetime


class Monitors(TypedDict):
    data: List[Monitor]


class MonitorStatus(TypedDict):
    createdAt: datetime
    lowerBound: float
    passed: bool
    predicted: float
    result: float
    type: Literal['decimal', 'integer']
    upperBound: float


class Metaplane(HttpClient):
    BASE_URL = 'https://dev.api.metaplane.dev'
    VERSION = 'v1'

    def __init__(self, config: Union[Dict, Config]):
        if type(config) is dict:
            self.config = Config.load(config=config)
        else:
            self.config = config

        self.headers = {
            'Authorization': f'Bearer {self.config.api_token}',
            'Content-Type': 'application/json',
        }

    def process(
        self,
        connection_id: str = None,
        monitor_ids: List[str] = None,
        poll_interval: int = 60,
        raise_on_failure: bool = False,
    ) -> None:
        print('Metaplane process started.')

        if not monitor_ids and connection_id:
            print(f'Fetching monitors for connection ID {connection_id}.')
            monitors = self.monitors(connection_id)['data']
            monitor_ids = [m['id'] for m in monitors]
            time.sleep(10)

        monitors_started = {}
        print(f'Running {len(monitor_ids)} monitor(s).')
        for monitor_id in monitor_ids:
            print(f'Running monitor ID {monitor_id}.')
            now = datetime.utcnow()
            status = self.run_monitors([monitor_id])['status']
            print(f'Monitor ID {monitor_id} ran with status {status}.')
            if 200 == status:
                monitors_started[monitor_id] = now

        monitors_completed = {}
        while len(monitors_completed) < len(monitors_started):
            for monitor_id in monitor_ids:
                print(f'Checking status for monitor ID {monitor_id}.')
                monitor_status = self.monitor_status(monitor_id)

                start_date = monitors_started[monitor_id]
                status_date = dateutil.parser.parse(monitor_status['createdAt'])
                completed = status_date.timestamp() >= start_date.timestamp()
                print(f'Monitor ID {monitor_id} completed: {completed}; started {start_date}, '
                      f'last created {status_date}).')
                if completed:
                    passed = monitor_status['passed']

                    if raise_on_failure and not passed:
                        raise Exception(f'Monitor ID {monitor_id} didn’t pass.')

                    monitors_completed[monitor_id] = passed
                    print(f'Monitor ID {monitor_id} passed: {passed}.')

            if len(monitors_completed) < len(monitors_started):
                time.sleep(poll_interval)

        print('Metaplane process completed.')

    def connections(self, include_deleted: bool = False) -> List[Connection]:
        return self.make_request(
            f'/{self.VERSION}/connections',
            params=dict(includeDeleted=include_deleted),
        )

    def monitors(self, connection_id: str, include_deleted: bool = False) -> Monitors:
        return self.make_request(
            f'/{self.VERSION}/monitors/connection/{connection_id}',
            params=dict(includeDeleted=include_deleted),
        )

    def monitor_status(self, monitor_id: str) -> MonitorStatus:
        return self.make_request(f'/{self.VERSION}/monitors/status/{monitor_id}')

    def run_monitors(self, monitor_ids: List[str]) -> Dict:
        response = self.make_request(f'/{self.VERSION}/monitors/run', method='POST', payload=dict(
            testIds=monitor_ids,
        ))
        return dict(status=response.status_code)



from mage_ai.streaming.sinks.amazon_s3 import AmazonS3Sink
from mage_ai.tests.base_test import TestCase
from unittest.mock import patch


class AmazonS3Tests(TestCase):
    def test_init(self):
        with patch.object(AmazonS3Sink, 'init_client') as mock_init_client:
            AmazonS3Sink(dict(
                connector_type='amazon_s3',
                bucket='bucket',
                prefix='prefix',
            ))
            mock_init_client.assert_called_once()

    def test_init_invalid_config(self):
        with patch.object(AmazonS3Sink, 'init_client') as mock_init_client:
            with self.assertRaises(Exception) as context:
                AmazonS3Sink(dict(
                    connector_type='amazon_s3',
                    bucket='bucket',
                ))
            self.assertTrue(
                '__init__() missing 1 required positional argument: \'prefix\''
                in str(context.exception),
            )
            self.assertEqual(mock_init_client.call_count, 0)



from mage_ai.api.presenters.BasePresenter import BasePresenter
import os


class FolderPresenter(BasePresenter):
    default_attributes = [
        'name',
        'path',
    ]

    def name(self, **kwargs) -> str:
        return os.path.basename(self.resource.path)



from mage_ai.data_preparation.repo_manager import get_repo_path
from mage_ai.io.azure_blob_storage import AzureBlobStorage
from mage_ai.io.config import ConfigFileLoader
from pandas import DataFrame
from os import path

if 'data_exporter' not in globals():
    from mage_ai.data_preparation.decorators import data_exporter


@data_exporter
def export_data_to_azure_blob_storage(df: DataFrame, **kwargs) -> None:
    """
    Template for exporting data to a Azure Blob Storage.
    Specify your configuration settings in 'io_config.yaml'.

    Docs: https://docs.mage.ai/design/data-loading
    """
    config_path = path.join(get_repo_path(), 'io_config.yaml')
    config_profile = 'default'

    container_name = 'your_container_name'
    blob_path = 'your_blob_path'

    AzureBlobStorage.with_config(ConfigFileLoader(config_path, config_profile)).export(
        df,
        container_name,
        blob_path,
    )



from mage_ai.api.presenters.BasePresenter import BasePresenter


class ClusterPresenter(BasePresenter):
    default_attributes = [
        'clusters',
        'id',
        'success',
        'type',
    ]



from google.cloud import run_v2
from google.iam.v1 import iam_policy_pb2, policy_pb2
from google.oauth2 import service_account
from googleapiclient.discovery import build
from typing import Any

import os
import time


class CloudRunServiceManager:
    def __init__(
        self,
        project_id: str,
        path_to_credentials_json_file: str,
        region='us-west2',
    ):
        self.project_id = project_id
        self.region = region

        scopes = [
            'https://www.googleapis.com/auth/cloud-platform',
            'https://www.googleapis.com/auth/compute',
        ]
        credentials = service_account.Credentials.from_service_account_file(
            path_to_credentials_json_file,
            scopes=scopes,
        )

        self.services_client = run_v2.ServicesClient(credentials=credentials)
        self.compute_service = build('compute', 'v1', credentials=credentials)

    def list_services(self) -> Any:
        response = self.services_client.list_services(
            run_v2.ListServicesRequest(
                parent=f'projects/{self.project_id}/locations/{self.region}'
            )
        )

        forwarding_rules = self.compute_service.globalForwardingRules().list(
            project=self.project_id
        ).execute()

        ip_map = dict()
        for rule in forwarding_rules.get('items', []):
            rule_name = rule.get('name')
            # Get service name by removing '-urlmap'
            service_name = rule_name.split('-urlmap')[0]
            ip_map[service_name] = rule.get('IPAddress')

        services = []
        for service in response:
            name = service.name.split('/')[-1]
            services.append(dict(
                ip=ip_map.get(name),
                name=name,
                status='RUNNING',
                type='run service',
            ))

        return services

    def create_service(self, service_id) -> None:
        resource_prefix = f'projects/{self.project_id}/locations/{self.region}'
        existing_service_name = os.getenv('GCP_SERVICE_NAME')
        existing_service = self.services_client.get_service(
            run_v2.GetServiceRequest(
                name=f'{resource_prefix}/services/{existing_service_name}'
            )
        )
        existing_service.name = None
        existing_service.template.revision = None
        existing_service.template.containers[0].command = None

        env_vars = existing_service.template.containers[0].env
        env_vars.append(
            run_v2.types.EnvVar(
                name='USER_CODE_PATH',
                value=service_id
            )
        )
        existing_service.template.containers[0].env = env_vars

        service_request = run_v2.CreateServiceRequest(
            parent=resource_prefix,
            service=existing_service,
            service_id=service_id,
        )

        self.services_client.create_service(service_request)

        resource = f'{resource_prefix}/resources/{service_id}'
        iam_request = iam_policy_pb2.GetIamPolicyRequest(resource=resource)
        policy = self.services_client.get_iam_policy(request=iam_request)

        policy.bindings.append(policy_pb2.Binding(role='roles/run.invoker', members=['allUsers']))

        set_iam_request = iam_policy_pb2.SetIamPolicyRequest(resource=resource, policy=policy)
        self.services_client.set_iam_policy(request=set_iam_request)

        self.__create_load_balancer(service_id)

    def __create_load_balancer(self, service_id) -> None:
        # Create NEG
        print('Creating network endpoint group...')
        neg_obj = {
            'name': f'{service_id}-neg',
            'networkEndpointType': 'SERVERLESS',
            'cloudRun': {
                'service': service_id
            }
        }

        neg_service = self.compute_service.regionNetworkEndpointGroups()
        neg_response = neg_service.insert(
            project=self.project_id,
            region=self.region,
            body=neg_obj
        ).execute()
        print('Network endpoint group created!')
        group_url = neg_response.get('targetLink')

        # Create backend service
        print('Creating backend service...')
        backend_service_name = f'{service_id}-urlmap-backend-default'
        backend_body = {
            'name': backend_service_name,
            'loadBalancingScheme': 'EXTERNAL',
            'backends': [{'group': group_url}],
            'enableCDN': False,
            # TODO: create a new security policy
            'securityPolicy': f"{os.getenv('GCP_SERVICE_NAME')}-security-policy",
            'iap': {
                'enable': False,
                'oauth2ClientId': '',
                'oauth2ClientSecret': '',
            },
            'logConfig': {
                'enable': True,
                'sampleRate': None
            }
        }

        backends_service = self.compute_service.backendServices()
        backends_response = backends_service.insert(
            project=self.project_id,
            body=backend_body
        ).execute()
        print('Backend service created!')
        backend_service_url = backends_response.get('targetLink')

        # Create external IP address
        print('Creating external IP address...')
        address_name = f'{service_id}-urlmap-address'
        addresses_service = self.compute_service.globalAddresses()
        addresses_service.insert(project=self.project_id, body={'name': address_name}).execute()

        ip_address = addresses_service.get(
            project=self.project_id,
            address=address_name
        ).execute().get('address')
        print('External IP address created!')

        # Create url map
        print('Creating url map...')
        url_map_body = {
            'name': f'{service_id}-urlmap-url-map',
            'defaultService': backend_service_url,
        }
        url_maps_service = self.compute_service.urlMaps()
        url_maps_response = self.__try_creating_resource(
            url_maps_service.insert(
                project=self.project_id,
                body=url_map_body
            ),
            'Url map'
        )

        url_map_link = url_maps_response.get('targetLink')

        # Create http proxy
        print('Creating target http proxy...')
        http_proxy_body = {
            'name': f'{service_id}-urlmap-http-proxy',
            'urlMap': url_map_link,
        }
        http_proxy_service = self.compute_service.targetHttpProxies()
        http_proxy_response = self.__try_creating_resource(
            http_proxy_service.insert(
                project=self.project_id,
                body=http_proxy_body
            ),
            'Target http proxy'
        )
        http_proxy_link = http_proxy_response.get('targetLink')

        # Create forwarding rules
        print('Creating forwarding rule...')
        forwarding_rules_body = {
            'name': f'{service_id}-urlmap',
            'IPAddress': ip_address,
            'IPProtocol': 'TCP',
            'portRange': '80-80',
            'target': http_proxy_link,
            'loadBalancingScheme': 'EXTERNAL'
        }

        forwarding_rules_service = self.compute_service.globalForwardingRules()
        self.__try_creating_resource(
            forwarding_rules_service.insert(
                project=self.project_id,
                body=forwarding_rules_body
            ),
            'Forwarding rule'
        )

    def __try_creating_resource(self, request, resource: str):
        attempts = 0
        while attempts < 10:
            attempts += 1
            try:
                response = request.execute()
                print(f'{resource} created!')
                return response
            except Exception:
                print(f'{resource}: still creating..., sleeping for 30 seconds')
                time.sleep(30)
        raise Exception(f'Creating {resource} timed out')



from mage_ai.data_integrations.sources.constants import SQL_SOURCES
from mage_ai.data_preparation.logging.logger import DictLogger
from mage_ai.data_preparation.models.pipelines.integration_pipeline import IntegrationPipeline
from mage_ai.orchestration.db import db_connection
from mage_ai.orchestration.db.models.schedules import BlockRun, PipelineRun
from mage_ai.orchestration.metrics.pipeline_run import calculate_metrics
from mage_ai.shared.array import find
from mage_ai.shared.hash import index_by, merge_dict
from typing import Dict, List
import json
import math
import os
import shutil


SQL_SOURCES_UUID = [d.get('uuid', d['name'].lower()) for d in SQL_SOURCES]


def clear_source_output_files(
    pipeline_run: PipelineRun,
    logger: DictLogger,
) -> None:
    tags = dict(
        pipeline_run_id=pipeline_run.id,
        pipeline_uuid=pipeline_run.pipeline_uuid,
    )

    integration_pipeline = IntegrationPipeline.get(pipeline_run.pipeline_uuid)

    for stream in integration_pipeline.streams():
        tap_stream_id = stream['tap_stream_id']
        source_output_folder = integration_pipeline.source_output_folder(tap_stream_id)
        if os.path.exists(source_output_folder):
            logger.info(
                f'Removing source output directory for stream {tap_stream_id} at '
                f'{source_output_folder}.',
                tags=merge_dict(tags, dict(
                    source_output_folder=source_output_folder,
                )),
            )
            shutil.rmtree(source_output_folder)


def initialize_state_and_runs(
    pipeline_run: PipelineRun,
    logger: DictLogger,
    variables: Dict,
) -> List[BlockRun]:
    tags = dict(
        pipeline_run_id=pipeline_run.id,
        pipeline_uuid=pipeline_run.pipeline_uuid,
    )

    try:
        update_stream_states(pipeline_run, logger, variables)

        block_runs = create_block_runs(pipeline_run, logger)

        logger.info(
            f'Calculate metrics for pipeline run {pipeline_run.id} started.',
            tags=tags,
        )
        calculate_metrics(pipeline_run)
        logger.info(
            f'Calculate metrics for pipeline run {pipeline_run.id} completed.',
            tags=merge_dict(tags, dict(metrics=pipeline_run.metrics)),
        )

        return block_runs
    except Exception as err:
        pipeline_run.update(status=PipelineRun.PipelineRunStatus.FAILED)

        logger.exception(
            str(err),
            error=err,
            tags=tags,
        )

        raise err


def create_block_runs(pipeline_run: PipelineRun, logger: DictLogger) -> List[BlockRun]:
    from mage_integrations.sources.constants import BATCH_FETCH_LIMIT

    integration_pipeline = IntegrationPipeline.get(pipeline_run.pipeline_uuid)

    blocks = integration_pipeline.get_executable_blocks()
    executable_blocks = []

    # we expect that blocks in integration pipelines only have 1 downstream block, and
    # that there is only 1 root block and 1 leaf block.
    current_block = integration_pipeline.data_loader
    while True:
        block = find(lambda b: b.uuid == current_block.uuid, blocks)
        executable_blocks.append(block)
        downstream_blocks = current_block.downstream_blocks
        if len(downstream_blocks) == 0:
            break
        current_block = downstream_blocks[0]

    data_loader_block = integration_pipeline.data_loader
    data_exporter_block = integration_pipeline.data_exporter

    transformer_blocks = [b for b in executable_blocks if b.uuid not in [
        data_loader_block.uuid,
        data_exporter_block.uuid,
    ]]

    is_sql_source = integration_pipeline.source_uuid in SQL_SOURCES_UUID
    record_counts_by_stream = {}
    if is_sql_source:
        record_counts_by_stream = index_by(lambda x: x['id'], integration_pipeline.count_records())

    arr = []

    for stream in integration_pipeline.streams():
        tap_stream_id = stream['tap_stream_id']

        tags = dict(
            source=integration_pipeline.source_uuid,
            stream=tap_stream_id,
        )

        record_counts = None
        if is_sql_source:
            record_counts = record_counts_by_stream[tap_stream_id]['count']
        number_of_batches = math.ceil((record_counts or 1) / BATCH_FETCH_LIMIT)
        tags2 = merge_dict(tags, dict(
            number_of_batches=number_of_batches,
            record_counts=record_counts,
        ))

        logger.info(
            f"Number of records for stream {tap_stream_id}: "
            f"{'N/A' if record_counts is None else record_counts}.",
            tags=tags2,
        )
        logger.info(
            f"Number of batches for loading data from stream {tap_stream_id}: {number_of_batches}.",
            tags=tags2,
        )

        for idx in range(number_of_batches):
            arr += [
                pipeline_run.create_block_run(
                    f'{data_loader_block.uuid}:{tap_stream_id}:{idx}',
                    commit=False,
                ),
            ] + [pipeline_run.create_block_run(
                    f'{b.uuid}:{tap_stream_id}:{idx}',
                    commit=False,
                ) for b in transformer_blocks] + [
                pipeline_run.create_block_run(
                    f'{data_exporter_block.uuid}:{tap_stream_id}:{idx}',
                    commit=False,
                ),
            ]
            try:
                db_connection.session.commit()
            except Exception as e:
                db_connection.session.rollback()
                raise Exception(f'Failed to create block runs for {tap_stream_id}:{idx}.') from e
    return arr


def update_stream_states(pipeline_run: PipelineRun, logger: DictLogger, variables: Dict) -> None:
    from mage_integrations.sources.utils import update_source_state_from_destination_state

    integration_pipeline = IntegrationPipeline.get(pipeline_run.pipeline_uuid)

    for stream in integration_pipeline.streams(variables):
        tap_stream_id = stream['tap_stream_id']
        destination_table = stream.get('destination_table', tap_stream_id)

        tags = dict(
            destination_table=destination_table,
            pipeline_run_id=pipeline_run.id,
            pipeline_uuid=pipeline_run.pipeline_uuid,
            stream=tap_stream_id,
        )

        logger.info(
            f'Updating source state from destination state for stream '
            f'{tap_stream_id} and table {destination_table}.',
            tags=tags,
        )
        source_state_file_path = integration_pipeline.source_state_file_path(
            destination_table=destination_table,
            stream=tap_stream_id,
        )
        destination_state_file_path = integration_pipeline.destination_state_file_path(
            destination_table=destination_table,
            stream=tap_stream_id,
        )
        update_source_state_from_destination_state(
            source_state_file_path,
            destination_state_file_path,
        )

        with open(source_state_file_path, 'r') as f:
            text = f.read()
            d = json.loads(text) if text else {}
            bookmark_values = d.get('bookmarks', {}).get(tap_stream_id)
            logger.info(
                f'Source state for stream {tap_stream_id}: {bookmark_values}',
                tags=merge_dict(tags, d),
            )

        with open(destination_state_file_path, 'r') as f:
            text = f.read()
            d = json.loads(text) if text else {}
            bookmark_values = d.get('bookmarks', {}).get(tap_stream_id)
            logger.info(
                f'Destination state for stream {tap_stream_id}: {bookmark_values}',
                tags=merge_dict(tags, d),
            )
